module 5-BinomialHeap where

  {- Courtesy of Josh Ko -}

-- A least-significant-bit-first representation
-- of binary numbers.
-- For example, decimal 11 = binary 1011
-- is represented as one∷ (one∷ (zero∷ (one∷ nul))).

open import Prelude

data Bin : Set where
  nul   : Bin
  zero∷ : Bin → Bin
  one∷  : Bin → Bin

  -- There are some issues with this representation.
  -- For example, 1 has many representions:
  --
  --    one∷ nul
  --    one∷ zero∷ nul
  --    one∷ zero∷ zero∷ nul
  --       ...
  --
  -- Later in this file will want to enforce that
  -- there is no zero∷ after the last one∷.
  -- For now, it suffices.

 -- How to increment a binary number by one?

incr : Bin → Bin
incr = {!   !}

2* : ℕ → ℕ
2* n = n + n

bin→ℕ : Bin → ℕ
bin→ℕ = {!   !}

  -- if you need 2* or addition, etc,
  -- define your own!

data Bool : Set where
  false : Bool
  true : Bool

data ⊤ : Set where
  tt : ⊤

data _×_ (A B : Set) : Set where
  _,_ : A → B → A × B

infixr 5 _×_ _,_

 {- Given X : ℕ → Set and n : ℕ,
    X ^ n is a sequence of n items, ending with ⊤.
      X n , X (n-1) , ... , X 1, ⊤.
 -}

_^_ : (ℕ → Set) → ℕ → Set
X ^ zero    = ⊤
X ^ (suc n) = X n × (X ^ n)

{-
  Example:

  X ^ 0 = ⊤
  X ^ 1 = X 0 × ⊤
  X ^ 2 = X 1 × X 0 × ⊤
  X ^ 3 = X 2 × X 1 × X 0 × ⊤
        :
-}

  -- Today, Vec is only used as an example,
  -- helping you to understand what X ^ n is.

ex1 : Vec Bool ^ 3
ex1 = {!   !}

  -- Binomial Tree.

data BTree : ℕ → Set where
  node : ∀ {r} → (x : ℕ)
         → (ts : BTree ^ r) → BTree r

   -- m ≤? n is true iff m is less than or equal to n.

_≤?_ : ℕ → ℕ → Bool
m ≤? n = {!   !}

  -- The function below attaches one BTree r to
  -- another BTree r, to form a BTree (suc r).
  -- The smaller root is chosen as the new root.

attach : ∀ {r} → BTree r → BTree r
               → BTree (suc r)
attach (node x ts) (node y us) with x ≤? y
... | false = node y (node x ts , us)
... | true  = node x (node y us , ts)

  -- Binomial Heap

data BHeap : ℕ → Bin → Set where
  nul   : ∀ {n} → BHeap n nul
  zero∷ : ∀ {n m} → BHeap (suc n) m
          → BHeap n (zero∷ m)
  one_∷ : ∀ {n m} → BTree n → BHeap (suc n) m
          → BHeap n (one∷ m)

 -- inserting a binominal tree into a heap.

ins : ∀ {n m} → BTree n → BHeap n m → BHeap n (incr m)
ins = {!   !}

-- Adding binary numbers

_∔_ : Bin → Bin → Bin
m ∔ n = {!   !}

join : ∀ {i m n} → BHeap i m → BHeap i n
       → BHeap i (m ∔ n)
join = {!   !}

-- decrement a binary number

  -- Positive b is an evidence that b is not nul.

data Positive : Bin → Set where
   zero∷ : ∀ {b} → Positive b → Positive (zero∷ b)
   one∷  : ∀ {b} → Positive (one∷ b)

-- decrement a binary number if it is not nul.

decr : (b : Bin) → Positive b → Bin
decr = {!   !}

-- extract the leftmost tree

extract : ∀ {n m} → (p : Positive m)
        → BHeap n m → (BTree n × BHeap n (decr m p))
extract = {!   !}

-- How do we implement delMin?

-- Unfortunately it is not as easy as the operations
-- above. So far, types has helped us to develop
-- programs --- given a type there is only one sensible
-- way the program can be written. In the case of delMin,
-- however, the structure of the program does not follow
-- the type. Thus we have to go the other way round,
-- convincing the type checker that our program is correct.
-- And this is not always easy.

-- In the completed file, we demonstrate one of the ways
-- to implement delMin. Typically, a formal proof of a
-- program could be 10 times as large as the program itself.
